@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2008, Aconex.  All Rights Reserved.
+ * Copyright (c) 2006-2010, Aconex.  All Rights Reserved.
  * Copyright (c) 2006, Ken McDonell.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -31,11 +31,60 @@
 #include <qwt_legend_item.h>
 #include <qwt_scale_draw.h>
 #include <qwt_scale_widget.h>
+#include <qwt_scale_engine.h>
 #include <qwt_text.h>
 #include <qwt_text_label.h>
 
 #define DESPERATE 0
 
+//
+// *Always* clamp minimum metric value at zero when positive -
+// preventing confusion when values silently change up towards
+// the maximum over time (for pmchart, our users are expecting
+// a constant zero baseline at all times, or so we're told).
+//
+class ValueScaleEngine : public QwtLinearScaleEngine
+{
+public:
+    ValueScaleEngine() : QwtLinearScaleEngine()
+    {
+	my.autoScale = true;
+	my.minimum = -1;
+	my.maximum = -1;
+    }
+
+    double minimum() const { return my.minimum; }
+    double maximum() const { return my.maximum; }
+    bool autoScale() const { return my.autoScale; }
+    void setAutoScale(bool autoScale) { my.autoScale = autoScale; }
+    void setScale(bool autoScale, double minimum, double maximum)
+    {
+	my.autoScale = autoScale;
+	my.minimum = minimum;
+	my.maximum = maximum;
+    }
+
+    virtual void autoScale(int maxSteps, double &minValue,
+			   double &maxValue, double &stepSize) const
+    {
+	if (my.autoScale) {
+	    if (minValue > 0)
+		minValue = 0.0;
+	} else {
+	    minValue = my.minimum;
+	    maxValue = my.maximum;
+	}
+	QwtLinearScaleEngine::autoScale(maxSteps, minValue, maxValue, stepSize);
+    }
+
+private:
+    struct {
+	bool autoScale;
+	double minimum;
+	double maximum;
+    } my;
+};
+
 Chart::Chart(Tab *chartTab, QWidget *parent) : QwtPlot(parent), Gadget()
 {
     Gadget::setWidget(this);
@@ -55,19 +104,18 @@ Chart::Chart(Tab *chartTab, QWidget *parent) : QwtPlot(parent), Gadget()
 	    SLOT(showCurve(QwtPlotItem *, bool)));
 
     // start with autoscale y axis
+    my.engine = new ValueScaleEngine();
     setAxisAutoScale(QwtPlot::yLeft);
+    setAxisScaleEngine(QwtPlot::yLeft, my.engine);
     setAxisFont(QwtPlot::yLeft, *globalFont);
 
     my.tab = chartTab;
     my.title = NULL;
-    my.autoScale = true;
     my.rateConvert = true;
     my.antiAliasing = true;
     my.style = NoStyle;
     my.scheme = QString::null;
     my.sequence = 0;
-    my.yMin = -1;
-    my.yMax = -1;
     my.picker = new QwtPlotPicker(QwtPlot::xBottom, QwtPlot::yLeft,
 			QwtPicker::PointSelection | QwtPicker::DragSelection,
 			QwtPlotPicker::CrossRubberBand, QwtPicker::AlwaysOff,
@@ -81,6 +129,8 @@ Chart::Chart(Tab *chartTab, QWidget *parent) : QwtPlot(parent), Gadget()
 			 SLOT(moved(const QwtDoublePoint &)));
 
     replot();
+
+    console->post("Chart::ctor complete(%p)", this);
 }
 
 Chart::~Chart()
@@ -95,6 +145,7 @@ Chart::~Chart()
 	    free(plot->plotData);
 	free(plot);
     }
+    delete my.picker;
 }
 
 void Chart::setCurrent(bool enable)
@@ -116,13 +167,13 @@ void Chart::setCurrent(bool enable)
     palette.setColor(QPalette::Active, QPalette::Text,
 		enable ? globalSettings.chartHighlight : QColor("black"));
     titleLabel()->setPalette(palette);
+
     sp = axisWidget(QwtPlot::yLeft);
     t = sp->title();
     t.setColor(enable ? globalSettings.chartHighlight : "black");
     sp->setTitle(t);
 }
 
-
 void Chart::preserveLiveData(int i, int oi)
 {
 #if DESPERATE
@@ -318,7 +369,9 @@ void Chart::redoScale(void)
     // the upper bound of the y-axis range (hBound()) drives the choice
     // of appropriate units scaling.
     //
-    if (my.autoScale && axisScaleDiv(QwtPlot::yLeft)->hBound() > 1000) {
+    if (my.engine->autoScale() &&
+	axisScaleDiv(QwtPlot::yLeft)->upperBound() > 1000) {
+	double scaled_max = axisScaleDiv(QwtPlot::yLeft)->upperBound();
 	if (my.units.dimSpace == 1) {
 	    switch (my.units.scaleSpace) {
 		case PM_SPACE_BYTE:
@@ -346,34 +399,63 @@ void Chart::redoScale(void)
 		    rescale = true;
 		    break;
 	    }
+	    if (rescale) {
+		// logic here depends on PM_SPACE_* values being consecutive
+		// integer values as the scale increases
+		scaled_max /= 1024;
+		while (scaled_max > 1000) {
+		    my.units.scaleSpace++;
+		    scaled_max /= 1024;
+		    if (my.units.scaleSpace == PM_SPACE_EBYTE) break;
+		}
+	    }
 	}
 	else if (my.units.dimTime == 1) {
 	    switch (my.units.scaleTime) {
 		case PM_TIME_NSEC:
 		    my.units.scaleTime = PM_TIME_USEC;
 		    rescale = true;
+		    scaled_max /= 1000;
 		    break;
 		case PM_TIME_USEC:
 		    my.units.scaleTime = PM_TIME_MSEC;
 		    rescale = true;
+		    scaled_max /= 1000;
 		    break;
 		case PM_TIME_MSEC:
 		    my.units.scaleTime = PM_TIME_SEC;
 		    rescale = true;
+		    scaled_max /= 1000;
 		    break;
 		case PM_TIME_SEC:
 		    my.units.scaleTime = PM_TIME_MIN;
 		    rescale = true;
+		    scaled_max /= 60;
 		    break;
 		case PM_TIME_MIN:
 		    my.units.scaleTime = PM_TIME_HOUR;
 		    rescale = true;
+		    scaled_max /= 60;
 		    break;
 	    }
+	    if (rescale) {
+		// logic here depends on PM_TIME* values being consecutive
+		// integer values as the scale increases
+		while (scaled_max > 1000) {
+		    my.units.scaleTime++;
+		    if (my.units.scaleTime <= PM_TIME_SEC)
+			scaled_max /= 1000;
+		    else
+			scaled_max /= 60;
+		    if (my.units.scaleTime == PM_TIME_HOUR) break;
+		}
+	    }
 	}
     }
 
-    if (rescale == false && my.autoScale && axisScaleDiv(QwtPlot::yLeft)->hBound() < 0.1) {
+    if (rescale == false && my.engine->autoScale() &&
+	axisScaleDiv(QwtPlot::yLeft)->upperBound() < 0.1) {
+	double scaled_max = axisScaleDiv(QwtPlot::yLeft)->upperBound();
 	if (my.units.dimSpace == 1) {
 	    switch (my.units.scaleSpace) {
 		case PM_SPACE_KBYTE:
@@ -401,30 +483,57 @@ void Chart::redoScale(void)
 		    rescale = true;
 		    break;
 	    }
+	    if (rescale) {
+		// logic here depends on PM_SPACE_* values being consecutive
+		// integer values (in reverse) as the scale decreases
+		scaled_max *= 1024;
+		while (scaled_max < 0.1) {
+		    my.units.scaleSpace--;
+		    scaled_max *= 1024;
+		    if (my.units.scaleSpace == PM_SPACE_BYTE) break;
+		}
+	    }
 	}
 	else if (my.units.dimTime == 1) {
 	    switch (my.units.scaleTime) {
 		case PM_TIME_USEC:
 		    my.units.scaleTime = PM_TIME_NSEC;
 		    rescale = true;
+		    scaled_max *= 1000;
 		    break;
 		case PM_TIME_MSEC:
 		    my.units.scaleTime = PM_TIME_USEC;
 		    rescale = true;
+		    scaled_max *= 1000;
 		    break;
 		case PM_TIME_SEC:
 		    my.units.scaleTime = PM_TIME_MSEC;
 		    rescale = true;
+		    scaled_max *= 1000;
 		    break;
 		case PM_TIME_MIN:
 		    my.units.scaleTime = PM_TIME_SEC;
 		    rescale = true;
+		    scaled_max *= 60;
 		    break;
 		case PM_TIME_HOUR:
 		    my.units.scaleTime = PM_TIME_MIN;
 		    rescale = true;
+		    scaled_max *= 60;
 		    break;
 	    }
+	    if (rescale) {
+		// logic here depends on PM_TIME* values being consecutive
+		// integer values (in reverse) as the scale decreases
+		while (scaled_max < 0.1) {
+		    my.units.scaleTime--;
+		    if (my.units.scaleTime < PM_TIME_SEC)
+			scaled_max *= 1000;
+		    else
+			scaled_max *= 60;
+		    if (my.units.scaleTime == PM_TIME_NSEC) break;
+		}
+	    }
 	}
     }
 
@@ -817,7 +926,7 @@ void Chart::setStroke(Plot *plot, Style style, QColor color)
 	    plot->curve->setBrush(QBrush(color, Qt::SolidPattern));
 	    plot->curve->setStyle(QwtPlotCurve::Sticks);
 	    if (my.style == UtilisationStyle)
-		setScale(true, my.yMin, my.yMax);
+		my.engine->setAutoScale(true);
 	    break;
 
 	case AreaStyle:
@@ -826,14 +935,14 @@ void Chart::setStroke(Plot *plot, Style style, QColor color)
 	    plot->curve->setStyle(isStepped(plot) ?
 				  QwtPlotCurve::Steps : QwtPlotCurve::Lines);
 	    if (my.style == UtilisationStyle)
-		setScale(true, my.yMin, my.yMax);
+		my.engine->setAutoScale(true);
 	    break;
 
 	case UtilisationStyle:
 	    plot->curve->setPen(QColor(Qt::black));
 	    plot->curve->setStyle(QwtPlotCurve::Steps);
 	    plot->curve->setBrush(QBrush(color, Qt::SolidPattern));
-	    setScale(false, 0.0, 100.0);
+	    my.engine->setScale(false, 0.0, 100.0);
 	    break;
 
 	case LineStyle:
@@ -842,7 +951,7 @@ void Chart::setStroke(Plot *plot, Style style, QColor color)
 	    plot->curve->setStyle(isStepped(plot) ?
 				  QwtPlotCurve::Steps : QwtPlotCurve::Lines);
 	    if (my.style == UtilisationStyle)
-		setScale(true, my.yMin, my.yMax);
+		my.engine->setAutoScale(true);
 	    break;
 
 	case StackStyle:
@@ -850,7 +959,7 @@ void Chart::setStroke(Plot *plot, Style style, QColor color)
 	    plot->curve->setBrush(QBrush(color, Qt::SolidPattern));
 	    plot->curve->setStyle(QwtPlotCurve::Steps);
 	    if (my.style == UtilisationStyle)
-		setScale(true, my.yMin, my.yMax);
+		my.engine->setAutoScale(true);
 	    break;
 
 	case NoStyle:
@@ -979,17 +1088,15 @@ void Chart::setLabel(int m, QString s)
 
 void Chart::scale(bool *autoScale, double *yMin, double *yMax)
 {
-    *autoScale = my.autoScale;
-    *yMin = my.yMin;
-    *yMax = my.yMax;
+    *autoScale = my.engine->autoScale();
+    *yMin = my.engine->minimum();
+    *yMax = my.engine->maximum();
 }
 
 void Chart::setScale(bool autoScale, double yMin, double yMax)
 {
-    my.yMin = yMin;
-    my.yMax = yMax;
-    my.autoScale = autoScale;
-    if (my.autoScale)
+    my.engine->setScale(autoScale, yMin, yMin);
+    if (autoScale)
 	setAxisAutoScale(QwtPlot::yLeft);
     else
 	setAxisScale(QwtPlot::yLeft, yMin, yMax);
@@ -1010,13 +1117,14 @@ void Chart::setRateConvert(bool rateConvert)
 void Chart::setYAxisTitle(const char *p)
 {
     QwtText *t;
+    bool enable = (my.tab->currentGadget() == this);
 
     if (!p || *p == '\0')
 	t = new QwtText(" ");	// for y-axis alignment (space is invisible)
     else
 	t = new QwtText(p);
     t->setFont(*globalFont);
-    t->setColor(globalSettings.chartHighlight);
+    t->setColor(enable ? globalSettings.chartHighlight : "black");
     setAxisTitle(QwtPlot::yLeft, *t);
 }
 
@@ -1090,14 +1198,19 @@ void Chart::save(FILE *f, bool hostDynamic)
     SaveViewDialog::saveChart(f, this, hostDynamic);
 }
 
-void Chart::print(QPainter *qp, QRect &rect)
+void Chart::print(QPainter *qp, QRect &rect, bool transparent)
 {
     QwtPlotPrintFilter filter;
 
-    filter.setOptions(QwtPlotPrintFilter::PrintAll &
-	~QwtPlotPrintFilter::PrintCanvasBackground &
-	~QwtPlotPrintFilter::PrintWidgetBackground &
-	~QwtPlotPrintFilter::PrintGrid);
+    if (transparent)
+	filter.setOptions(QwtPlotPrintFilter::PrintAll &
+	    ~QwtPlotPrintFilter::PrintBackground &
+	    ~QwtPlotPrintFilter::PrintGrid);
+    else
+	filter.setOptions(QwtPlotPrintFilter::PrintAll &
+	    ~QwtPlotPrintFilter::PrintGrid);
+
+    console->post("Chart::print: options=%d", filter.options());
     QwtPlot::print(qp, rect, filter);
 }
 
@@ -1173,20 +1286,29 @@ void Chart::setupTree(QTreeWidget *tree)
 void Chart::addToTree(QTreeWidget *treeview, QString metric,
 	const QmcContext *context, bool isInst, QColor &color, QString &label)
 {
-    QRegExp regex(tr("\\.|\\[|\\]"));
+    QRegExp regexInstance("\\[(.*)\\]$");
+    QRegExp regexNameNode(tr("\\."));
     QString source = context->source().source();
-    QStringList	baselist;
+    QString inst, name = metric;
+    QStringList	namelist;
     int depth;
 
     console->post("Chart::addToTree src=%s metric=%s, isInst=%d",
 		(const char *)source.toAscii(), (const char *)metric.toAscii(),
 		isInst);
 
-    baselist = metric.split(regex);
-    baselist.prepend(source);	// add the host/archive root as well.
-    depth = baselist.size();
-    if (baselist.at(depth-1).isEmpty())	// regex side-effect on ']'.
-	depth--;
+    depth = name.indexOf(regexInstance);
+    if (depth > 0) {
+	inst = name.mid(depth+1);	// after '['
+	inst.chop(1);			// final ']'
+	name = name.mid(0, depth);	// prior '['
+    }
+
+    namelist = name.split(regexNameNode);
+    namelist.prepend(source);	// add the host/archive root as well.
+    if (depth > 0)
+	namelist.append(inst);
+    depth = namelist.size();
 
     // Walk through each component of this name, creating them in the
     // target tree (if not there already), right down to the leaf.
@@ -1195,7 +1317,7 @@ void Chart::addToTree(QTreeWidget *treeview, QString metric,
     NameSpace *item = NULL;
 
     for (int b = 0; b < depth; b++) {
-	QString text = baselist.at(b);
+	QString text = namelist.at(b);
 	bool foundMatchingName = false;
 	for (int i = 0; i < tree->childCount(); i++) {
 	    item = (NameSpace *)tree->child(i);
